/*!
\file errorHandler.c
\author Robert Sperling <sperling@small-window-manager.de>
\version 1.3.6
\brief error handling system 

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
\endverbatim 
*/

#ifdef TRACE
#define ERRORMSG
char traceTree [30][30];
#endif

#ifndef MINIMAL
#define ignoreXErrors(statement) (\
	XSync(display,0), XSetErrorHandler(expectedXErrorHandler),	\
	statement,	\
	XSetErrorHandler(xErrorHandler))
#endif

#ifdef ERRORMSG
#include <stdarg.h>

#ifndef MINIMAL
#ifdef DEBUG
static int debuglevel = 0;
#endif//DEBUG
static int nodebug    = 0;
#endif//MINIMAL

// -err-------------------------------------------------------------------
//! prints an error message to stderr
/*!
\verbatim
   v1.3.0-pre1: initial release
\endverbatim
*/
void err(char *fmt, ...)
{
//  int counter;
  va_list argp;

  if (!nodebug)
  { 
//    for (counter=0 ; counter<debuglevel ; counter++) 
//      fprintf(stderr," ");

    fprintf(stdout, "  sWM: > ");
    va_start(argp, fmt);
    vfprintf(stdout, fmt, argp);
    va_end(argp);
    fprintf(stdout, "\n");
  }// end if nodebug 
}//err
#endif//ERRORMSG





#ifdef TRACE
// -trace---------------------------------------------------------------
//! shows, in which functions sWM is in
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.6-pre4: added smarter traceing system
\endverbatim
*/
void trace(int level, char *fmt, ...)
{
#ifdef SHOWTRACE
  int counter;
#endif//SHOWTRACE  
  va_list argp;

  if (!nodebug)
  { if (level<0) 
    { debuglevel=debuglevel+level;
      sprintf(traceTree[debuglevel+1]," ");    
      sprintf(traceTree[debuglevel+2]," ");    
    }
    if (debuglevel<0) 
    { fprintf(stderr,"########### panic! Trace tree underrun. ###########\n");
      debuglevel=0;
    }
/*#ifdef SHOWTRACE
    if (level<0)
    { for (counter=0 ; counter<debuglevel; counter++) 
        fprintf(stderr," ");
      fprintf(stderr, "SWM: ");
    }  
#endif
*/  
    if (level>=0) 
    { debuglevel=debuglevel+level;
      va_start(argp, fmt);
      if (level==0) vsprintf(traceTree[debuglevel+1], fmt, argp);
      else          vsprintf(traceTree[debuglevel], fmt, argp);
      va_end(argp);
/*#ifdef SHOWTRACE
      fprintf (stderr,"entered: ");
#endif */
    }
/*#ifdef SHOWTRACE
    else
    { fprintf (stderr,"leaving: ");
    }
      va_start(argp, fmt);
      vfprintf(stderr, fmt, argp);
      va_end(argp);
      fprintf(stderr, "\n");
#endif    
*/
#ifdef SHOWTRACE
  fprintf(stdout,"sWMtrace: main");
  for (counter=1; counter<debuglevel+2; counter++)
  { fprintf(stdout,"->%s",traceTree[counter]);
  }
  fprintf(stdout,"\n");
#endif
  }//if !nodebug     
}//trace
#endif//TRACE












#ifdef VERBOSE
// -showEvents--------------------------------------------------------
//! displays the event received
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre4: added some Events
\endverbatim   
*/
void showEvents(XEvent e)
{ char *s;
  Window w;
  switch (e.type) 
  {
  // Mouse 
  case ButtonPress:
  s = "ButtonPress"; w = e.xbutton.window; break;
  case ButtonRelease:
  s = "ButtonRelease"; w = e.xbutton.window; break;
  case MotionNotify:
  s = "MotionNotify"; w = e.xmotion.window; break;
  // Keyboard 
  case KeyPress:
#ifdef KEYS			
  s = "KeyEvent"; w = e.xkey.window; 
#endif//KEYS
       break;
  case KeyRelease:
#ifdef KEYS			
  s = "KeyRelease"; w = e.xkey.window; 
#endif//KEYS
       break;
  case KeymapNotify:
#ifdef KEYS			
  s = "KeymapNotify"; w = e.xkey.window; 
#endif//KEYS
       break;
  case ClientMessage:
  s = "ClientMessage"; w = e.xclient.window; break;
  case ColormapNotify:
  s = "ColormapNotify"; w = e.xcolormap.window; break;
  case ConfigureNotify:
  s = "ConfigNotify"; w = e.xconfigure.window; break;
  case ConfigureRequest:
  s = "ConfigRequest"; w = e.xconfigurerequest.window; break;
  case CreateNotify:
  s = "CreateNotify"; w = e.xcreatewindow.window; break;
  case DestroyNotify:
  s = "DestroyNotify"; w = e.xdestroywindow.window; break;
  // Window change 
  case EnterNotify:
  s = "EnterNotify"; w = e.xcrossing.window; break;
  case LeaveNotify:
  s = "*unhandled LeaveNotify"; w = e.xcrossing.window; break;
  case FocusIn:
  s = "FocusIn"; w = e.xfocus.window; break;
  case FocusOut:
  s = "FocusOut"; w = e.xfocus.window; break;
  case Expose:
  s = "ExposeEvent"; w = e.xexpose.window; break;
  case GraphicsExpose:
  s = "GraphicsExposeEvent (unhandled)"; w = e.xexpose.window; break;
  case NoExpose:
  s = "NoExposeEvent (unhandled)"; w = e.xnoexpose.drawable; break;
  case MapNotify:
  s = "MapNotify"; w = e.xmap.window; break;
  case MapRequest:
  s = "MapRequest"; w = e.xmaprequest.window; break;
  case MappingNotify:
  s = "MappingNotify"; w = e.xmapping.window; break;
  case VisibilityNotify:
  s = "VisibilityNotify"; w = e.xvisibility.window; break;
  case GravityNotify:
  s = "GravityNotify"; w = e.xgravity.window; break;
  case PropertyNotify:
  s = "PropertyNotify"; w = e.xproperty.window; break;
  case ReparentNotify:
  s = "ReparentNotify"; w = e.xreparent.window; break;
  case ResizeRequest:
  s = "ResizeRequest"; w = e.xresizerequest.window; break;
  case UnmapNotify:
  s = "UnmapNotify"; w = e.xunmap.window; break;
  case CirculateNotify:
  s = "CirculateNotify" ; w = e.xcirculate.window; break;
  case CirculateRequest:
  s = "CirculateRequest" ; w = e.xcirculate.window; break;
  default:
  s = "unknown event"; w = None; break;
}// end showEvents 

  swmerr ("received %s,\twindow = %#x", s, w);
}//showEvents 
#endif//VERBOSE






#ifndef MINIMAL
// -debug_getWmState------------------------------------------------
//! returns the windows wm state in a safe way
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.6:      TESTING XGetWindowProperty with unsigned char*
\endverbatim   
*/
int debug_getWmState(Client *client)
{ Atom real_type;
  int real_format, state;
  unsigned long n, extra;
  unsigned char *data;
  int returnState;
 
  ignoreXErrors(returnState=XGetWindowProperty(display, client->window, 
                                 xa_wm_state, 0L, 2L, False,
                                 AnyPropertyType, &real_type, 
				 &real_format, &n, &extra, &data));
  if (returnState == Success && n) 
  { state = *(int *)data;
    if (data) XFree(data);
    return state;
  }// end if XGetWindowProperty
  
  return WithdrawnState;
}//debug_getWmState
#endif//debug_getWmState



#ifdef DEBUG
// -dumpClients--------------------------------------------------------
//! dumps sWMs client database to stderr and shows the trace tree
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.6-pre4: now showing also the trace tree, this function can
                also be run by SIGUSR1
   v1.3.6:      dumping clients now to stdout
\endverbatim   
*/
void dumpClients()
{ Client *client;
  int counter=0;

#ifdef TRACE
  fprintf(stdout,"traceTree: main");
  for (counter=0; counter<debuglevel+2; counter++)
  { fprintf(stdout,"%s->",traceTree[counter]);
  }
  fprintf(stdout,"end\n");
#endif

  counter=0;

  fprintf (stdout,"rootwin is: (%#lx)\n",rootWindow);

  XGrabServer(display);  

  fprintf(stdout,"*****************************************************************\n");
  client = headClient;
  while (client)
  { counter++;
    if (client->name) fprintf (stdout,"*%s\n",client->name);
    else fprintf(stdout,"*unknown window/name\n");

    fprintf(stdout," state:%d,",debug_getWmState(client));
    fprintf(stdout,"iUnmaps:%d,",client->ignoreUnmap);

    if (client->window) fprintf (stdout,"win: (%#lx),",client->window);
    else fprintf(stdout,"nowindow,");
    if (client->parent) fprintf (stdout,"par: (%#lx),",client->parent);
    else fprintf (stdout,"noparent,");
#ifdef LAYERS
    nodebug++;
    fprintf(stdout," layer:%ld",getValuedHint(client->window,xa_windowLayer));
    fprintf(stdout," curlayer:%d",client->layer);
    nodebug--;
#endif
//    fprintf(stdout," %#x (%#x) @ %d,%d", 
//	client->x, client->y);
   fprintf(stdout,"\n");
   client=client->next;
  }	
  fprintf(stdout,"*****************************************************************\n");
  XUngrabServer(display);
  fprintf(stdout,"ClientCounter  : %d\n",clientCounter);
  fprintf(stdout,"Clients in List: %d\n",counter);
}//dumpClients
#endif








// ****************************************************************************
//  xErrorHandler
//! should handle all sWM-internal errors
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.6:      now showing X-Errors on stderr
\endverbatim
*/
int xErrorHandler(Display *display, XErrorEvent *event)
{ 
  Client 	*client;
#ifdef ERRORMSG
  char 		errorMessage[80];
#endif//ERRORMSG

#ifdef TRACE
  int counter;
  
  while (xerrorHandlerRunning);
  
  xerrorHandlerRunning=true;
    
  fprintf(stderr,"!!!!!!!!!!!!!!!!!!!!!! xErrorHandler !!!!!!!!!!!!!!!!!!!!!!\n");
  trace(1,"xErrorHandler");
  fprintf(stderr,"sWMtrace: main");

  for (counter=1; counter<debuglevel+1; counter++)
  { fprintf(stderr,"->%s",traceTree[counter]);
  }
  fprintf(stderr,"\n");
#endif//TRACE

  if (event->error_code == BadAccess && event->resourceid == rootWindow) 
  { 
#ifdef STD_IO
    fprintf(stderr,"root window unavailible (maybe another wm is running?)\n");
#endif//STD_IO

#ifndef NOROOTCHECK
    exit(1);
#endif//NOROOTCHECK
  }//if rootWin unavailable 
  else
  { 
#ifdef ERRORMSG
    XGetErrorText(display, event->error_code, errorMessage, sizeof(errorMessage));
    fprintf(stderr,"X-ERROR (%#x): %s (code: %i) in request %#x\n", 
            (int)event->resourceid, errorMessage, (int)event->error_code,
	     event->request_code);
    nodebug++;
#endif//ERRORMSG      

    // OK, trying to find out who caused the trouble...
    client=getClient(event->resourceid,anyWindow);
#ifdef ERRORMSG
    nodebug--;
#endif//ERRORMSG        

    if (client)
    { 
#ifdef NAMEHANDLING
      if (client->name) 
      { 
#ifdef ERRORMSG
        fprintf(stderr,"ERROR: caused by client: %s, window: %#lx\n",
	        client->name,event->resourceid);
#endif//ERRORMSG
      }
#endif//NAMEHANDLING 

/* Ich hoffe dies wird bald nicht mehr notwendig sein... */
 if ((event->resourceid==client->window) &&
     (event->error_code==BadWindow)) 
 {
#ifdef VERBOSE 
   swmerr ("before remove client, state of removeParent is: %i...\n",
       removeParentCurrentWindow);
#endif//VERBOSE

#ifdef ERRORMSG       
   if (removeParentCurrentWindow)
     swmerr ("ERROR removeParent running - hmm, this should not happen");
   else
#endif//ERRORMSG   
   {
#ifdef GNOME
#ifndef NORAISE   
     if (client==raisedClient) raisedClient=headClient;
#endif//NORAISE     
#endif//GNOME
//     removeParent(client, false);
     clientToRemove=client;
   }

#ifdef VERBOSE   
   swmerr ("after remove client...\n");
#endif//VERBOSE
    }
   }//if client 
  }//else if rootwin 

#ifdef TRACE
  fprintf(stderr,
          "!!!!!!!!!!!!!!!!!!! end of xErrorHandler !!!!!!!!!!!!!!!!!!!\n");
  trace(-1,"xErrorHandler");
#endif//TRACE
  xerrorHandlerRunning=false;

  return 0;
}//xErrorHandler








// **********************************************************************
#ifndef MINIMAL
//  cleanUpErrorHandler
//! is only here for debugging - I hate this, really
/*!
\verbatim
   v1.3.0-pre1: initial release
\endverbatim
\todo find a way to ensure a window is still valid
*/
int cleanUpErrorHandler(Display *display, XErrorEvent *event)
{ 
#ifdef ERRORMSG
  char errorMessage[80];
#endif//ERRORMSG
  Client *client;
  
#ifdef TRACE
  trace(1,"cleanUpErrorHandler");
#endif//TRACE
  
  
  client=getClient(event->resourceid,anyWindow);

#ifdef ERRORMSG
  XGetErrorText(display, event->error_code, errorMessage, sizeof(errorMessage));
  fprintf(stderr,"cleanup X-Error (%#lx): %s", event->resourceid, errorMessage);
  if (client && client->name) fprintf(stderr,"* client: %s",client->name);  

  XGetErrorText(display, event->request_code, errorMessage, 
                sizeof(errorMessage));
  fprintf(stderr,"request: %s", errorMessage);

  if (client && client->window==event->resourceid) fprintf(stderr,"error in client");

#ifdef PIXMAP
  if (client && client->closebox==event->resourceid) fprintf(stderr,"error in closebox");
#ifdef MAXIMIZE
  if (client && client->maxbox==event->resourceid) fprintf(stderr,"error in maxbox");
#endif//MAXIMIZE

#ifdef GNOME
  if (client && client->stickybox==event->resourceid) fprintf(stderr,"error in stickybox");
#endif//GNOME  
#endif//PIXMAPS
#endif//ERRORMSG

  if (client && event->error_code == BadWindow) removeParent(client,false);

#ifdef TRACE
  trace(-1,"cleanUpErrorHandler");
#endif//TRACE

  return 0;
}// end cleanUpErrorHandler
#endif // NOT MINIMAL






// **********************************************************************
//  expectedErrorHandler
//! carefully ignores some errors
/*! This functions is important for handling unmanaged windows
\verbatim
   v1.3.0-pre1: initial release
\endverbatim
\todo find a way to ensure a window is still valid
*/
int expectedXErrorHandler(Display *display, XErrorEvent *event)
{ 
#ifdef DEBUG	
  char errorMessage[80];
#endif
  
#ifdef DEBUG
  XGetErrorText(display, event->error_code, errorMessage, 
                sizeof(errorMessage));

#ifdef VERBOSE
  XGetErrorText(display, event->request_code, errorMessage, 
                sizeof(errorMessage));
  swmerr ("* cleanup error (%#x): %s", event->resourceid, errorMessage);
#endif//VERBOSE
  
#endif
  return 0;
}//expectedErrorHandler

