/*! 
\file gnome.c
\version 1.3.6
\author Robert Sperling <sperling@small-window-manager.de>
\brief GNOME compliance code and hint management

This file contains all functions to make sWM GNOME and MWM compliant

\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         * 
 ***************************************************************************
\endverbatim
*/

#ifdef GNOME
Atom			   xa_windowHints;
Atom			   xa_windowState;
Atom 			   xa_windowClientList;
Atom 			   xa_windowProtocols;
Atom			   xa_windowDeskButtonProxy;
Atom 			   xa_wm_desktop;
Atom			   xa_wm_maxDesktops;
Atom			   xa_AtomList[20];
Atom			   xa_windowSupportingWmCheck;
Atom			   xa_windowProtocols;

#ifdef EWMH
Atom                       xa_netSupported;  
Atom			   xa_netNumberOfDesktops;
Atom			   xa_netCurrentDesktop;
Atom			   xa_netDesktopViewport;

// -initEWMHCompliance--------------------------------------------
//! initializes all nessesary atoms for enhanced WM hints compliance
/*!
\verbatim
   v1.3.6:      initial release
\endverbatim
*/
void initEWMHCompliance()
{ int                    count = 0;
  unsigned long 	 data[4];
  unsigned int maxDesktops = NUMBEROFDESKTOPS;
  
#ifdef TRACE
  trace (1,"initEWMHCompliance");
#endif//TRACE

  xa_netNumberOfDesktops = XInternAtom(display, "_NET_NUMBER_OF_DESKTOPS", False);
  xa_AtomList[count++]	    = xa_netNumberOfDesktops;
  XChangeProperty(display, rootWindow, xa_netNumberOfDesktops, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&maxDesktops, 1);


  xa_netCurrentDesktop   = XInternAtom(display, "_NET_CURRENT_DESKTOP", False);
  xa_AtomList[count++]	    = xa_netCurrentDesktop;
  XChangeProperty(display, rootWindow, xa_netCurrentDesktop, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&currentDesktop, 1);


  xa_netDesktopViewport  = XInternAtom(display, "_NET_DESKTOP_VIEWPORT", False);
  xa_AtomList[count++]	    = xa_netDesktopViewport;
  data[0]=0;
  data[1]=0;
  XChangeProperty(display, rootWindow, xa_netDesktopViewport, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&data, 2);


  xa_netSupported        = XInternAtom(display, "_NET_SUPPORTED", False);
  xa_AtomList[count++]	    = xa_netSupported;
  XChangeProperty(display, rootWindow, xa_netSupported, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&xa_AtomList, 1);

#ifdef TRACE
  trace (-1,"initEWMHCompliance");
#endif//TRACE
}//initEWMHCompliance
#endif//EWMH

// -initGnomeCompliance--------------------------------------------
//! initializes all nessesary atoms and the gnome required window
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: fixed trace at the end
   v1.3.6:      added sWM-Panel
\endverbatim
*/
void initGnomeCompliance()
{ int count=0;
  unsigned int maxDesktops = NUMBEROFDESKTOPS;
#ifdef DESKCHANGESCRIPT
  char toExec[255];
#endif//DESKCHANGESCRIPT

#ifdef SWMPANEL
  XSetWindowAttributes   setAttr;
#ifdef PIXMAPS
  int			 counter;
#endif//PIXMAPS
#endif//SWMPANEL

  
#ifdef TRACE
  trace (1,"initGnomeCompliance");
#endif//TRACE
  
  currentDesktop=0;
  xa_windowState	    = XInternAtom(display, "_WIN_STATE", False);
  xa_AtomList[count++]	    = xa_windowState;
  xa_windowHints	    = XInternAtom(display, "_WIN_HINTS", False);
  xa_AtomList[count++]	    = xa_windowHints;
  xa_windowClientList	    = XInternAtom(display, "_WIN_CLIENT_LIST", False);
  xa_AtomList[count++]	    = xa_windowClientList;
  xa_windowDeskButtonProxy  = XInternAtom(display, "_WIN_DESKTOP_BUTTON_PROXY", False);
  xa_AtomList[count++]	    = xa_windowDeskButtonProxy;
  xa_wm_desktop		    = XInternAtom(display, "_WIN_WORKSPACE", False);
  xa_AtomList[count++]	    = xa_wm_desktop;
  xa_wm_maxDesktops	    = XInternAtom(display, "_WIN_WORKSPACE_COUNT", False);
  xa_AtomList[count++]	    = xa_wm_maxDesktops;

#ifdef LAYERS
  // Tell all clients that we support layers
  xa_AtomList[count++]	    = xa_windowLayer;
#endif//LAYERS


#ifdef SWMPANEL
  // Create sWM-PANEL window used as GNOME required window 
  setAttr.override_redirect 	= True;
  setAttr.background_pixel 	= sWMPanelColor.pixel;
  setAttr.border_pixel 		= borderColor.pixel; 
  setAttr.event_mask 		= ExposureMask | EnterWindowMask | ButtonPressMask;

  sWMWindow = XCreateWindow(display, rootWindow, 
    			    0, ymax-SWMPANELSIZE, xmax, SWMPANELSIZE, 
	 	            0, DefaultDepth(display, screen), 
			    CopyFromParent, DefaultVisual(display, screen),
                            CWOverrideRedirect | CWBackPixel | CWBorderPixel | CWEventMask,
			    &setAttr);						  

  XMapWindow(display,sWMWindow);
  XRaiseWindow(display,sWMWindow);

#ifdef PIXMAPS
  for (counter=0; counter<NUMBEROFSTARTERS; counter++)
  { 
    setAttr.background_pixmap   = starterPix[counter].pixmap;

    // no buttonPressMask makes panel clicks go through the window
    setAttr.event_mask 		= ExposureMask | EnterWindowMask;
    
    starterWindow[counter]= XCreateWindow(display, sWMWindow, 
    			    counter*SWMPANELSIZE, 0, SWMPANELSIZE, SWMPANELSIZE, 
	 	            0, DefaultDepth(display, screen), 
			    CopyFromParent, DefaultVisual(display, screen),
                            CWOverrideRedirect | CWBorderPixel | CWEventMask | CWBackPixmap,
			    &setAttr);						  

    XMapWindow(display,starterWindow[counter]);
    if (starterPix[counter].pixmap)  XFreePixmap (display, starterPix[counter].pixmap); 
  } 
#endif//PIXMAPS

  refreshPanel();
#else//SWMPANEL
  // Create GNOME required window 
  sWMWindow		    = XCreateSimpleWindow(display, rootWindow,
                                                  -200, -200, 5, 5, 0, 0, 0);
#endif//SWMPANEL
  
  // Set up GNOME properties 
  xa_windowSupportingWmCheck = XInternAtom(display, "_WIN_SUPPORTING_WM_CHECK", 
                                           False);

  XChangeProperty(display, sWMWindow, xa_windowSupportingWmCheck, XA_CARDINAL, 
                  32, PropModeReplace, (unsigned char*) &sWMWindow, 1);
  XChangeProperty(display, rootWindow, xa_windowSupportingWmCheck, XA_CARDINAL,
                  32, PropModeReplace, (unsigned char*) &sWMWindow, 1);

  XChangeProperty(display, sWMWindow, xa_windowDeskButtonProxy, XA_CARDINAL, 
                  32, PropModeReplace, (unsigned char*) &sWMWindow, 1);
  XChangeProperty(display, rootWindow, xa_windowDeskButtonProxy, XA_CARDINAL, 
                  32, PropModeReplace, (unsigned char*) &sWMWindow, 1);
  xa_windowProtocols	    = XInternAtom(display, "_WIN_PROTOCOLS", False);
  XChangeProperty(display, rootWindow, xa_windowProtocols, XA_ATOM, 32, 
                  PropModeReplace, (unsigned char*) xa_AtomList, count);
  XChangeProperty(display, rootWindow, xa_wm_maxDesktops, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&maxDesktops, 1);
  XChangeProperty(display, rootWindow, xa_wm_desktop, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&currentDesktop, 1);




  XSelectInput(display,rootWindow, SubstructureNotifyMask | 
                                   SubstructureRedirectMask |
				   ButtonPressMask |
				   ButtonReleaseMask |
				   FocusChangeMask |
				   PropertyChangeMask);

#ifdef DESKCHANGESCRIPT
      sprintf(toExec,"%s/.sWM/desktop%i &", CONFDIR, 1);
      system (toExec);
#endif//DESKCHANGESCRIPT

				
#ifdef TRACE
  trace (-1,"initGnomeCompliance");
#endif//TARCE
}//initGnomeCompliance



// -changeDesktop---------------------------------------------------------
//! switches to another Desktop keeping all sticky windows
/*!
\verbatim
   v1.3.0-pre1: initial release   
   v1.3.5:      now setting WIN_STATE_HID_WORKSPACE
   v1.3.6-pre3:	fixed: sticky windows were not shown with Alt-Cycle-Key,
   		fixed nasty bug ( == in assignment of desktop)
   v1.3.6-pre4: reordered to ensure proper trace
\endverbatim   
*/
void changeDesktop (unsigned int newDesktop)
{ Client *client;
#ifdef DESKCHANGESCRIPT
  char toExec[255];
#endif//DESKCHANGESCRIPT

#ifdef TRACE
  trace(1,"changeDesktop");
#endif//TRACE

  // merely here for speedup
  if (newDesktop!=currentDesktop) 
  { if (newDesktop<NUMBEROFDESKTOPS) 
    { 
#ifdef VERBOSE
      swmerr ("changing from desktop %d to %d\n", currentDesktop+1, newDesktop+1);
#endif//VERBOSE

#ifdef DESKCHANGESCRIPT
      sprintf(toExec,"%s/.sWM/desktop%i &", CONFDIR, newDesktop+1);
      system (toExec);
#endif//DESKCHANGESCRIPT

      client=headClient;
      while(client)
      {
#ifdef CHECKCLIENTS
        if (client->hints!=getValuedHint(client->window,xa_windowHints))
	{ 
#ifdef ERRORMSG	
	  if (client->name)
	    swmerr ("ERROR: client->hints is wrong on client: %s!",client->name);
          else	    
	    swmerr ("ERROR: client->hints is wrong on unnamed client!");
#endif//ERRORMSG	  
          client->hints=getValuedHint(client->window,xa_windowHints);
        }
#endif//CHECKCLIENTS
        if (client->sticky)
        { setValuedHint(client->window, xa_wm_desktop, newDesktop); 
          client->desktop=newDesktop;
        }
        else
        { if(getValuedHint(client->window,xa_windowState)!=IconicState)
          { if (client->desktop==newDesktop)
              unhideClient(client);
            else
              hideClient(client);
          }//client is shown 
	  toggleHint(client->window,xa_windowHints,WIN_STATE_HID_WORKSPACE);
        }//client is not sticky 
        client=client->next;
      }//while client 
      currentDesktop=newDesktop;

      // tell GNOME the new desktop
      setValuedHint(rootWindow, xa_wm_desktop, currentDesktop);
#ifdef EWMH      
      setValuedHint(rootWindow, xa_netCurrentDesktop, currentDesktop);
#endif//EWMH      
    }
    else//newDesktop<NUMBEROFDESKTOPS 
    { 
#ifdef ERRORMSG
      swmerr ("ERROR: new desktop number exceeds number of desktops");
#endif//ERRORMSG  
    }//newdesktop>numberofdesktops
  }//currentdesktop!=newdesktop
  else
  { 
#ifdef ERRORMSG  
    swmerr ("ERROR: currentDesktop==newDesktop");
#endif//ERRORMSG
  }
#ifdef TRACE
  trace(-1,"changeDesktop");
#endif//TRACE
}//changeDesktop 


// -updateClientList-----------------------------------------------
//! updates client list of gnome
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: changed error message to fprintf, err will now be used for 
                debuging only
   v1.3.5:      now also updating cached version of windowHints
   v1.3.6-pre2: added cast to remove compile error in malloc
   v1.3.6-pre4: now updating cached version of windowHints only when needed
   v1.3.6:	added per desktop client counter
\endverbatim   
*/ 
void updateClientList() 
{ 
  Client *client; 
  CARD32 *windowList; 
  long counter; 
 
#ifdef TRACE 
  trace(1,"updateClientList"); 
#endif//TRACE

#ifdef VERBOSE
  swmerr ("got %d clients",clientCounter);
  dumpClients(); 
  XSync(display,false);
#endif//VERBOSE 

  if ((windowList=(CARD32 *)malloc(sizeof(CARD32)*clientCounter))==NULL) 
  {
#ifdef STD_IO
    fprintf(stderr,"could not allocate memory");
#endif//STD_IO
  }     
  else
  {
#ifdef CHECKCLIENTS  
    if (!headClient) swmerr ("headClient not valid");
#ifdef VERBOSE    
    else swmerr ("headClient is valid");
#endif//VERBOSE  
#endif//CHECKCLIENTS    

    client=headClient;
    counter=0; 

    while (client)
    { if (client->window) 
      { 
#ifdef CHECKCLIENTS      
        if (client->hints!=getValuedHint(client->window,xa_windowHints))
	{ client->hints=getValuedHint(client->window,xa_windowHints);
#ifdef ERRORMSG
	  if (client->name) 
	    swmerr ("ERROR: client->hints is wrong on client: %s!",client->name);
	  else
	    swmerr ("ERROR: client->hints is wrong on unknown client!");
#endif//ERRORMSG                
	}  
#endif//CHECKCLIENTS
        windowList[counter]=client->window; 
	
//#ifndef PANELSHOWALL
//	clientDesktopCounter[client->desktop]++;
//#endif//PANELSHOWALL
	
      }//if client->window
      counter++; // here, because there might be windows just appearing
      client=client->next;
    } 
    
    if (clientCounter!=counter)
    { clientCounter=counter;
#ifdef ERRORMSG
      swmerr ("ERROR: client counter differs from known client windows (FIXME)");
#ifdef DEBUGPARANOID
      exit(-1);
#endif//DEBUGPARANOID
#endif//ERRORMSG
      updateClientList();
    }//client!=counter
    else
    { XChangeProperty(display, rootWindow, xa_windowClientList, XA_CARDINAL, 32, 
                      PropModeReplace, (unsigned char *)windowList, clientCounter); 
    }
    if(windowList) free(windowList); 
  }//else if no memory could be allocated

#ifdef TRACE 
  trace(-1,"updateClientList"); 
#endif//TRACE
}//updateClientList 

/***************************************************************************/

//  moveClientToDesktop
//! moves a client to another desktop and switches to it
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.6-pre4: new check if the client is already on the current desktop
\endverbatim   
*/ 
void moveClientToDesktop(Client *client, unsigned int newDesktop) 
{ 
#ifdef TRACE 
  trace (1,"moveClientToDesktop");
#endif//TRACE
 
  if (client->desktop!=newDesktop)
  { setValuedHint(client->window, xa_wm_desktop, newDesktop);
    client->desktop=newDesktop; 
    if(newDesktop!=currentDesktop) changeDesktop(newDesktop);
  } 
  else
  { 
#ifdef ERRORMSG  
    swmerr ("client is already on this desktop");
#endif//ERRORMSG    
  }
#ifdef TRACE 
  trace (-1,"moveClientToDesktop");
#endif//TRACE
}//moveClientToDesktop 

/***************************************************************************/

//  toggleClientSticky 
//! make client follow desktop switch (or not)
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: fixed unsticky, client is now left on current desktop
   v1.3.0-pre3: updated pixmap support
   v1.3.6:	no not updating client's desktop anymore
\endverbatim   
*/ 
void toggleClientSticky(Client *client) 
{ 
#ifdef TRACE
  trace (1,"toggleClientSticky"); 
#endif 

  toggleHint(client->window, xa_windowState, WIN_STATE_STICKY); 

// FIXME, REMOVEME this should not be nessesary...
  client->desktop=currentDesktop;
  
  if (client->sticky) 
  { client->sticky=false;

#ifdef PIXMAPS      
    // change sticky-button to unsticky
    XSetWindowBackgroundPixmap (display, client->stickybox,
	                        unStickyBoxPix.pixmap);
#endif	
  }    
  else
  { client->sticky=true;

#ifdef PIXMAPS      
    // change sticky-button to sticky again
    XSetWindowBackgroundPixmap (display, client->stickybox,
	                        stickyBoxPix.pixmap);
#endif	
  }// end else if already sticky
  redrawParent(client);
  
#ifdef TRACE
  trace (-1,"toggleClientSticky"); 
#endif 
}// end toggleClientSticky 

/***************************************************************************/

#ifndef NORAISE
//  raiseNextClient
//! raises the next client on local desktop
/*!
\verbatim
   v1.3.0-pre3: initial release
   v1.3.0-pre4: fixed a bug loosing raised clients, gnome workaround for
                non-layer support
		current desktop		
\endverbatim		
*/
void raiseNextClient() 
{ 
  Client *client;

#ifdef TRACE 
  trace(0,"raiseNextClient(*)"); 
#endif 

// Client *getNextClient(int keepOnDesktop, int skipFocus, int skipWinList);
 client=getNextClient(true,false,true);

#ifdef VERBOSE
 swmerr ("finished getNextClient");
#endif
// TRASH
//#ifdef LAYERS
//  if (!client) client=getBoundingClient(0,true);
//#else  // LAYERS 
//  if (!client) client=headClient;
//#endif // LAYERS 

 if (client) raiseClient (client);
}// end raiseNextClient 

/***************************************************************************/

//  getNextClient
//! gets the next client with some restrictions
/*!
\verbatim
   v1.3.0-pre1: no implementation
   v1.3.0-pre2: no implementation
   v1.3.0-pre3: no implementation
   v1.3.0-pre4: initial release
   v1.3.1:      fixed crash when item is last in list
   v1.3.3:	fixed little bug cycling up windows that were not on the
	        current desktop
\endverbatim
*/ 
Client *getNextClient(int keepOnDesktop, int skipFocus, int skipWinList)
{ Client *client;
  int     finished=false;
#ifdef TRACE
  trace(1,"getNextClient");
#endif

#ifndef NORAISE
  client = raisedClient;
  if (client && client->next) client = client->next; 
  else 
#endif//NORAISE  
    client=headClient;

#ifdef VERBOSE
  if (client->name) swmerr ("starting with %s",client->name);
#endif//VERBOSE

  while (!finished)
  { finished=true; 

    if (client && client->window)
    {
#ifdef VERBOSE
      swmerr ("now before keepOnDesktop");
#endif//VERBOSE
      if (keepOnDesktop && (client->desktop!=currentDesktop)) finished=false;

#ifdef VERBOSE
      swmerr ("now before skipFocus");
#endif//VERBOSE
      if (skipFocus && (client->hints & WIN_HINTS_SKIP_FOCUS)) 
        finished=false;

#ifdef VERBOSE
      swmerr ("now before skipWinList");
#endif//VERBOSE
      if (skipWinList && (getValuedHint(client->window,xa_windowHints)
                          &WIN_HINTS_SKIP_WINLIST)) finished=false;

//      if  case 3: if (!client->hints & WIN_HINTS_SKIP_TASKBAR) finished=true;
      if (!finished) 
      { 
        if (client->next) client=client->next;
	else client=headClient;
#ifdef VERBOSE
        if (client->name) swmerr ("next was %s",client->name);
#endif//VERBOSE
        if (client==raisedClient) 
	{ finished=true;
#ifdef VERBOSE
          if (client->name) swmerr ("raising %s",client->name);
#endif//VERBOSE
	  return NULL; // not dangerous to raise a client again
	}//client==raisedClient
      }//!finished
    }//if client  
    else 
    { client=headClient;
#ifdef VERBOSE
      if (client->name)
        swmerr ("now changing to headclient %s",client->name);
#endif//VERBOSE
    }
  }//while !finished

#ifdef TRACE
  trace(-1,"getNextClient");
#endif//TRACE
  return client;
}//getNextClient 
#endif//NORAISE 
#endif//GNOME

/***************************************************************************
 * Layer-Support                                                           *
 ***************************************************************************/

#ifdef LAYERS
//  setLayer
//! set the new layer for the client
/*! Gnome compliance also includes layer support 
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: no change
   v1.3.0-pre3: no change
   v1.3.0-pre4: no change
   v1.3.0-pre5: for testing, now uses adjustlayer
   v1.3.1:      new Layer management via raise/lower, eliminated oldlayer
   v1.3.2:      out of range message
\endverbatim   
\todo FOR TESTING: does adjustLayer work properly here?
*/ 
void setLayer(Client *client, int newLayer) 
{
#ifdef TRACE 
  trace (1,"setLayer"); 
#endif //TRACE

#ifdef VERBOSE
  if (client->name) swmerr ("setting client: %s to layer %d",client->name,newLayer);
  if (client->layer == newLayer) swmerr ("no change, not setting layer");
#endif//VERBOSE

  if (newLayer == client->layer) return; 
  
  if ((newLayer<0) || (newLayer>13))  
  {
#ifdef ERRORMSG
    swmerr ("error: layer %d is out of range",newLayer);
#ifdef DEBUGPARANOID
    exit(-1);
#endif//DEBUGPARANOID    
#endif//ERRORMSG
    return;
  }//if layer out of range
  setValuedHint(client->window, xa_windowLayer, newLayer); 
  adjustLayer(client,(newLayer>client->layer)); 

#ifdef TRACE 
  trace (-1,"setLayer"); 
#endif//TRACE

}//setLayer 

/***************************************************************************/

//  getBoundingClient 
//! returns the top most client in this layer
/*!
\returns top most client (topClient=1), or the bottom most (topClient = 0)

\verbatim   
   v1.3.0-pre2: initial release
   v1.3.0-pre4: eliminated getLayer (too simple)
   v1.3.0-pre5: sWM is now VERY much more carefull with (maybe) destroyed
                windows and syncs before querying window list 
   v1.3.1:      a bit of fine tuning now checking getClient only if it is
                visible, now grabbing server, using client->layer instead of
		asking X, fixed very stupid bug: break has to be IN if clause
		in bottonLayer
   v1.3.2:	fixed a debug level problem
\endverbatim
*/
Client *getBoundingClient(int givenLayer, int topClient)
{ Window *windowList, w1, w2; 
  int max, counter; 
  Client *client,*outer=NULL; 
  XWindowAttributes attr; 

#ifdef TRACE
  trace (1,"getBoundingClient");
#endif

  XGrabServer(display);
  XQueryTree(display, rootWindow, &w1, &w2, &windowList, &max); 

  for(counter=0; counter<max; counter++) 
  { 
    XGetWindowAttributes(display, windowList[counter], &attr);
    if (attr.map_state == IsViewable)
    { 
#ifdef DEBUG  
#ifndef DEBUGPARANOID  
      nodebug++;
#endif//DEBUGPARANOID
#endif//DEBUG
      client=getClient(windowList[counter],parentWindow); 
#ifdef DEBUG  
#ifndef DEBUGPARANOID  
      nodebug--;
#endif//DEBUGPARANOID
#endif//DEBUG    
      if(client) 
      { if (topClient) 
        { if(client->layer<=givenLayer) outer=client; 
          else break; 
        }
        else
        { if(client->layer>=givenLayer) 
	  { outer=client; 
	    break;
	  } 
        }// end else if top
      }// end if client  
    }// end if isviewable 
  }// end for 
  if(windowList) XFree(windowList); 

  XUngrabServer(display);

#ifdef DEBUG
  if (!outer) swmerr ("returning NULL");
#endif
 
#ifdef TRACE
  trace (-1,"getBoundingClient");
#endif

 return outer; 

}// end getBoundingClient 

// ***************************************************************************

//  adjustLayer
//! updated window stacking mode
/*!
\verbatim
   v1.3.0-pre2: initial release
   v1.3.0-pre4: eliminated getLayer (too simple)
   v1.3.1:      now own sibling management
\endverbatim
\todo FOR TESTING: new sibling management code (stable?)
*/
void adjustLayer (Client *client, int stackMode)
{ Client *sibling=NULL; 
  XWindowChanges wc;
  int clientLayer = WIN_LAYER_NORMAL;

#ifdef TRACE
  trace (1,"adjustLayer");
#endif//TRACE

#ifdef CHECKCLIENTS
  if (client)
  {
#endif//CHECKCLIENTS
    clientLayer=getValuedHint(client->window, xa_windowLayer);

#ifdef VERBOSE
    swmerr ("adjusting in Layer: %d",clientLayer);
#endif//VERBOSE
    switch(stackMode) 
    { case Above: 
      // Window wants to be raised to top of its layer 
        sibling=getBoundingClient(clientLayer,true); 
	
        // If there are no other windows in this layer, say it is the 
	// bottom most, so every new window will appear above this one 
        if(!sibling) 
        { wc.stack_mode=Below; 
#ifdef VERBOSE
          swmerr ("got no sibling windows, lowering this one");
#endif//VERBOSE
        }
        else 
        { wc.stack_mode=Above;
#ifdef VERBOSE
          if (client->name && sibling->name)
	    swmerr ("got sibling: %s for client: %s",sibling->name,client->name);
#endif//VERBOSE
	}
        break; 
      case Below: 
      // Window wants to be lowered to bottom of its layer 
        sibling=getBoundingClient(clientLayer,false); 
        wc.stack_mode=Below;
        break; 
    }//switch 

    if (sibling) 
    { if (sibling==client)
      {
#ifdef VERBOSE
        swmerr ("sibling is client itself, doing nothing");
#endif//VERBOSE
      }
      else 
      { 
#ifdef VERBOSE
        swmerr ("restacking using sibling");
#endif//VERBOSE
        wc.sibling=sibling->parent;
        // It is not enough to simply configure the parent. Otherwise
	// after a gnome-panel switch the next window clicked will no be rised
        XConfigureWindow(display,client->parent,CWSibling|CWStackMode,&wc);
//	wc.stack_mode=Above;
//        XConfigureWindow(display,client->window,CWStackMode,&wc);
      }
    } else
    { 
#ifdef VERBOSE
        swmerr ("restacking without sibling");
#endif//VERBOSE
      XConfigureWindow(display,client->parent,CWStackMode,&wc);
//      wc.stack_mode=Above;
//      XConfigureWindow(display,client->window,CWStackMode,&wc);
    }
//      XConfigureWindow(display,client->window,0,&wc);

#ifdef CHECKCLIENTS
  }//else client 
#endif//CHECKCLIENTS

 client->layer=clientLayer;
#ifdef TRACE
  trace (-1,"adjustLayer");
#endif//TRACE
}//adjustLayer 
#endif//LAYERS 

// ***************************************************************************
// * Hint-Management                                                         *
// ***************************************************************************

//  getValuedHint
//! reads a specific hint out of the list and returns its value
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre2: added layer support
   v1.3.0-pre3: added check for valid window
   v1.3.3:	more verbose now
   v1.3.6:	TESTING XGetWindowProperty with unsigned char*
\endverbatim
\warning even nessessary in minimal mode
*/
long getValuedHint(Window givenWin, Atom givenAtom) 
{ Atom realType; 
  int realFormat; 
  unsigned long readItems, itemsLeft; 
  long value=0;
  unsigned char *data;
#ifdef VERBOSE
  char *atomName; 
#endif//VERBOSE
		 
#ifdef TRACE 
  trace (0,"getValuedHint(*)"); 
#endif//TRACE 

#ifdef LAYERS
  // If the Application does not set its window layers, return our normal
  if (givenAtom==xa_windowLayer) value=windowLayerNormal;
#endif//LAYERS

#ifdef VERBOSE
  atomName=XGetAtomName(display, givenAtom);
  if (atomName)
    swmerr ("ATOM: %ld, Name:%s\n",givenAtom,atomName);
  else
    swmerr ("ATOM: %ld, Name unknown\n",givenAtom);
#endif//VERBOSE

#ifdef CHECKCLIENTS
  if (givenWin)
  { 
#endif//CHECKCLIENTS  

//    XSetErrorHandler(expectedXErrorHandler);
    if (XGetWindowProperty(display, givenWin, givenAtom, 0L, 1L, False, 
                           XA_CARDINAL,&realType, &realFormat, &readItems, 
  	  		   &itemsLeft,(unsigned char **)&data)==Success 
			   && readItems) 
    { value=*data; 
      XFree(data); 
    }
//    XSetErrorHandler(xErrorHandler);

#ifdef CHECKCLIENTS
  }
  else
  {
    swmerr ("window is not valid");
  }    
#endif//CHECKCLIENTS

  return value; 
}//getValuedHint

// ****************************************************************************

//  setValuedHint
//! sets a specific hint
/*!
\verbatim
   v1.3.0-pre1: initial release
   v1.3.0-pre3: added check for valid window
\endverbatim
\warning even in minimal mode neccesary
*/
void setValuedHint(Window givenWin, Atom givenAtom, long value) 
{ 
#ifdef TRACE
  trace(1,"setValuedHint"); 
#endif 

#ifdef CHECKCLIENTS
  if (givenWin)
  { 
#endif//CHECKCLIENTS  
    XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                    PropModeReplace, (unsigned char *)&value, 1); 
#ifdef CHECKCLIENTS
  }
  else
  {
    swmerr ("window is not valid");
  }		    
#endif//CHECKCLIENTS  

#ifdef TRACE
  trace(-1,"setValuedHint"); 
#endif//TRACE 

}// end setValuedHint

/****************************************************************************/

#ifdef GNOME
//  setHint
//! allows to set a bit in a hint value
/*!
\verbatim
   v1.3.0-pre1: initial release   
   v1.3.0-pre3: added window check
   v1.3.3:	now being more verbose
\endverbatim
*/
void setHint(Window givenWin, Atom givenAtom, long value) 
{ long newValue;

#ifdef TRACE
  trace(1,"setHint"); 
#endif 

#ifdef DEBUG
  if (givenWin)
  { 
#endif  
    newValue = getValuedHint (givenWin, givenAtom)|value;
  	   
#ifdef VERBOSE
    ignoreXErrors(XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&newValue, 1)); 
#else    
    XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                    PropModeReplace, (unsigned char *)&newValue, 1); 
#endif

#ifdef DEBUG
  }
  else
  {
    swmerr ("window is not valid");
  }		    
#endif  

#ifdef TRACE
  trace(-1,"setHint"); 
#endif 

}// end setHint

/****************************************************************************/

//  toggleHint
//! allows to invert a bit in a hint
/*!
\verbatim
   v1.3.0-pre1: initial release   
   v1.3.0-pre3: added window check
   v1.3.3:      now being more verbose
\endverbatim   
*/
void toggleHint(Window givenWin, Atom givenAtom, long value) 
{ long newValue;

#ifdef TRACE
  trace(1,"toggleHint"); 
#endif 

  if (givenWin)
  { newValue = getValuedHint (givenWin, givenAtom)^value;

#ifdef VERBOSE
    ignoreXErrors(XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&newValue, 1)); 
#else  	   
    XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                    PropModeReplace, (unsigned char *)&newValue, 1); 
#endif
  }
  else
  {
#ifdef DEBUG
    swmerr ("window is not valid");
#endif  
  }		    

#ifdef TRACE
  trace(-1,"toggleHint"); 
#endif 

}// end toggleHint  

/****************************************************************************/

//  deleteHint
//! allows to unset a bit in a hint
/*!
\verbatim
   v1.3.0-pre1: initial release   
   v1.3.0-pre3: updated window check
   v1.3.3:      now being more verbose
\endverbatim
*/
void deleteHint(Window givenWin, Atom givenAtom, long value) 
{ long newValue;

#ifdef TRACE
  trace(1,"deleteHint"); 
#endif 

  if (givenWin)
  { newValue = getValuedHint (givenWin, givenAtom)&~value;
 
#ifdef VERBOSE  	   
    ignoreXErrors(XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                  PropModeReplace, (unsigned char *)&newValue, 1));

#else
    XChangeProperty(display, givenWin, givenAtom, XA_CARDINAL, 32, 
                    PropModeReplace, (unsigned char *)&newValue, 1);
#endif    
    
  }
  else
  {
#ifdef DEBUG
    swmerr ("window is not valid");
#endif  
  }	    

#ifdef TRACE
  trace(-1,"deleteHint"); 
#endif 

}// end deleteHint  
#endif // GNOME

// ***************************************************************************
