/*!
\file sWMPanel.c
\author Robert Sperling <sperling@small-window-manager.de>
\version 1.3.6
\brief everything needed to display and manage the swmPanel
The swmPanel needs GNOME compliance enabled because it uses the sWMWindow.
\verbatim
 ***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************
\endverbatim 
*/
#define DESKSWITCHBORDER 	SWMPANELSIZE/2
#define DESKSWITCHMIDDLE 	SWMPANELSIZE/4
#define STARTERSAREA 		SWMPANELSIZE*NUMBEROFSTARTERS
#define SWMMAXBUTTONLENGTH 	150
#ifdef SWMPANELCLOCK
#define SWMCLOCKSIZE 		50
#else
#define SWMCLOCKSIZE 		0
#endif//SWMPANELCLOCK

#ifdef SWMPANELTASKBAR
int sWMClientButtonLength = 0;

//! Displays all managed clients on the taskbar
void showClients()
{ Client *client;
  int counter=0;
  int numberOfChars;
  int textSize;

  XClearArea  (display, sWMWindow, 
               STARTERSAREA, 1,
               xmax-STARTERSAREA-SWMCLOCKSIZE, SWMPANELSIZE-1, false);
    
  if (clientCounter>0)
  { sWMClientButtonLength=(xmax-STARTERSAREA-SWMCLOCKSIZE-DESKSWITCHBORDER)/clientCounter;
#ifdef VERBOSE
    swmerr ("number of clients: %i, buttonlength: %i\n",clientCounter,sWMClientButtonLength);
#endif//VERBOSE

    if (sWMClientButtonLength>SWMMAXBUTTONLENGTH) 
      sWMClientButtonLength=SWMMAXBUTTONLENGTH;  

  

    for (client=headClient; client!=NULL; client=client->next)
    { 
      XDrawLine(display,sWMWindow,panelGC,
                STARTERSAREA+counter*sWMClientButtonLength,0,
                STARTERSAREA+counter*sWMClientButtonLength,SWMPANELSIZE);

      if (client->name)
      { numberOfChars=strlen(client->name);

        textSize=XTextWidth(sWMFont,client->name, numberOfChars);

	if (textSize>sWMClientButtonLength)
	  numberOfChars=numberOfChars*sWMClientButtonLength/textSize-1;

        XDrawString (display, sWMWindow, panelGC,
  	             3+STARTERSAREA+counter*sWMClientButtonLength, 
		     DESKSWITCHBORDER+sWMFont->ascent/2,
		     client->name, numberOfChars);
#ifdef SHOWFOCUS
	if (client==lastFocusedClient)
	{
           XDrawLine (display, sWMWindow, panelGC,
  	                STARTERSAREA+counter*sWMClientButtonLength, 1,
		        STARTERSAREA+(counter+1)*sWMClientButtonLength, 1);
           XDrawLine (display, sWMWindow, panelGC,
  	                STARTERSAREA+counter*sWMClientButtonLength+1, 2,
		        STARTERSAREA+counter*sWMClientButtonLength+1, SWMPANELSIZE+1);
	}		
#endif//SHOWFOCUS	
      }	   
      counter++;		 
    }  
    XDrawLine(display,sWMWindow,panelGC,
              STARTERSAREA+counter*sWMClientButtonLength,0,
              STARTERSAREA+counter*sWMClientButtonLength,SWMPANELSIZE+1);
  }		 
}
#endif//SWMPANELTASKBAR




//! redraws the panel-outline
void refreshPanel()
{ //DesktopSwitcher |
  XDrawLine(display, sWMWindow, 
	     borderGC, xmax-DESKSWITCHBORDER, 0, 
	               xmax-DESKSWITCHBORDER, SWMPANELSIZE);
  //DesktopSwitcher -
  XDrawLine(display, sWMWindow, 
	     borderGC, xmax- DESKSWITCHBORDER, DESKSWITCHBORDER, 
	               xmax, DESKSWITCHBORDER);

  //DesktopSwitcher [-]
  XDrawLine(display, sWMWindow, 
	     borderGC, xmax-DESKSWITCHMIDDLE-2, DESKSWITCHMIDDLE, 
	               xmax-DESKSWITCHMIDDLE+3, DESKSWITCHMIDDLE);

  //DesktopSwitcher [+]
  XDrawLine(display, sWMWindow, 
	     borderGC, xmax-DESKSWITCHMIDDLE-2, 3*DESKSWITCHMIDDLE, 
	               xmax-DESKSWITCHMIDDLE+3, 3*DESKSWITCHMIDDLE);
  XDrawLine(display, sWMWindow, 
	     borderGC, xmax-DESKSWITCHMIDDLE, 3*DESKSWITCHMIDDLE-2, 
	               xmax-DESKSWITCHMIDDLE, 3*DESKSWITCHMIDDLE+3);

#ifdef SWMPANELTASKBAR
  showClients();
#endif//SWMPANELTASKBAR

#ifdef SWMPANELCLOCK		       
  refreshPanelClock();		
#endif//SWMPANELCLOCK         

  //TopLine
  XDrawRectangle(display, sWMWindow, 
	     borderGC, 0, 0, xmax-1, SWMPANELSIZE);
  		       
}//refreshPanel



#ifdef NORAISE
Client *lastClient;
#endif//NORAISE

#ifdef SWMPANELHIDE
int    panelHidden = 0;
//! hides/unhides the panel
void togglePanelHidden()
{ if (panelHidden) 
  { panelHidden=0;
    XMoveResizeWindow(display, sWMWindow, 
                      0, ymax-SWMPANELSIZE, xmax, SWMPANELSIZE);
  }
  else
  { panelHidden=1;
    XMoveResizeWindow(display, sWMWindow, 
                      0, ymax-1, xmax, 1);
  }
}  
#endif//SWMPANELHIDE  

//! checks, what was clicked on the panel
void panelClick(int xpos, int ypos)
{ 
#ifdef VERBOSE
  swmerr ("sWMPanel-Click at position: %i,%i\n",xpos,ypos);
#endif//VERBOSE


  if (xpos>xmax-DESKSWITCHBORDER) 
  { if (ypos>DESKSWITCHBORDER) 
    { if (currentDesktop==0) changeDesktop(NUMBEROFDESKTOPS-1);
      else changeDesktop (currentDesktop-1);
    } else 
    { if (currentDesktop>NUMBEROFDESKTOPS-2) changeDesktop(0);
      else changeDesktop (currentDesktop+1);
    }
  } else if (xpos<STARTERSAREA) 
  { char toExec[255];
#ifdef VERBOSE  
    fprintf(stderr,"pressed starter # %i\n",xpos/SWMPANELSIZE+1);
#endif//VERBOSE    
    sprintf(toExec,"%s/.sWM/starters/starter%i &",CONFDIR,xpos/SWMPANELSIZE+1);
#ifdef VERBOSE
    fprintf(stderr,"starting: %s\n",toExec);
#endif//VERBOSE    
    system(toExec); 
  }
#ifdef SWMPANELTASKBAR
  else if (xpos>STARTERSAREA && clientCounter>0)
  { Client *client;
    int counter;
    int clickedClient=(xpos-STARTERSAREA)/sWMClientButtonLength;
    if (clickedClient<clientCounter)
    { client=headClient;
#ifdef VERBOSE
      swmerr ("clicked client #%i", clickedClient);
#endif//VERBOSE      
      for (counter=1; counter<=clickedClient; counter++)
         client=client->next;


#ifdef GNOME      
#ifndef NORAISE
      if (client==raisedClient)
#else
      if (client==lastClient)
#endif//NORAISE      
      {  
#ifdef SHADE      
         toggleClientShaded(client);
#endif//SHADE	 
      }
      else	
#endif//GNOME	
      { if (client->desktop!=currentDesktop)
          changeDesktop(client->desktop);
        raiseClient(client);
      }	

#ifdef NORAISE
      lastClient=client;
#endif//NORAISE
	
#ifdef VERBOSE
      if (client->name) fprintf(stderr,"clicked: %s\n",client->name);  
#endif//VERBOSE
    }
  }//SWMPANELTASKBAR
#endif//SWMPANELTASKBAR  
}//panelClick




#ifdef SWMPANELCLOCK
//! redraws the panel clock
void refreshPanelClock ()
{
	char *timeString;
	time_t now;

#ifdef VERBOSE
      fprintf(stderr,"refreshing panel clock\n");  
#endif//VERBOSE

	now = time (0);
	timeString = ctime (&now) + 11;
	XClearArea (display,sWMWindow,
	            xmax - SWMCLOCKSIZE, 1,
		    SWMCLOCKSIZE-DESKSWITCHBORDER, SWMPANELSIZE-1,false);
	XDrawString (display, sWMWindow, panelGC,
	             xmax - SWMPANELSIZE*2, DESKSWITCHBORDER+sWMFont->ascent/2,
		     timeString, 5);
}//refreshPanelClock


//! sets the alarm for the next panel redraw
void setPanelClockTimer(int delay)
{                                                                                    
  struct itimerval value;
  value.it_interval.tv_usec = 0;
  value.it_interval.tv_sec = 0;
  value.it_value.tv_usec = 0;
 
  if (delay)
   value.it_value.tv_sec = 60;                                                
  else 
  { time_t currentTime;
    struct tm * now;
    
    time(&currentTime);
    now=localtime(&currentTime);
    value.it_value.tv_sec = 60 - now->tm_sec;
  }
  
#ifdef VERBOSE
      fprintf(stderr,"setting new timer for panel clock\n");  
#endif//VERBOSE

  setitimer(ITIMER_REAL,&value,NULL);                                         
}//setPanelClockTimer
#endif//SWMPANELCLOCK
